#ifndef PRIVATE_RENDERING_H
#define PRIVATE_RENDERING_H

#include "../rendering.h"
#include "../textures.h"
#include "../util.h"

#include <SDL2/SDL.h>
#include <glad/glad.h>

#include <stdbool.h>

struct sprite_primitive {
    t_frect rect;
    t_color color;
    float rotation;
    t_texture_key texture_key;
    bool flip_x;
    bool flip_y;
    bool blend; /* must be explicitly stated, textures having alpha channel isn't enough */
};

/* interleaved vertex array data */
/* TODO: use int16_t for uvs */
/* TODO: int16_t could be used for positioning, but we would need to have more CPU calcs */
struct sprite_primitive_payload {
    /* upper-left */
    t_fvec2 v0;
    t_fvec2 uv0;
    t_color c0;
    /* bottom-left */
    t_fvec2 v1;
    t_fvec2 uv1;
    t_color c1;
    /* bottom-right */
    t_fvec2 v2;
    t_fvec2 uv2;
    t_color c2;
    /* upper-right */
    t_fvec2 v3;
    t_fvec2 uv3;
    t_color c3;
};

struct rect_primitive {
    t_frect rect;
    t_color color;
};

struct circle_primitive {
    float radius;
    t_color color;
    t_fvec2 position;
};

enum primitive_2d_type {
    PRIMITIVE_2D_SPRITE,
    PRIMITIVE_2D_RECT,
    PRIMITIVE_2D_CIRCLE,
};

struct primitive_2d {
    enum primitive_2d_type type;

    union {
        struct sprite_primitive sprite;
        struct rect_primitive rect;
        struct circle_primitive circle;
    };
};

/* union for in-place recalculation of texture coordinates */
union uncolored_space_triangle {
    /* pending for sending, uvs are not final as texture atlases could update */
    struct uncolored_space_triangle_primitive {
        t_fvec3 v0;
        t_fvec2 uv0; /* in pixels */
        t_fvec3 v1;
        t_fvec2 uv1; /* in pixels */
        t_fvec3 v2;
        t_fvec2 uv2; /* in pixels */
    } primitive;

    /* structure that is passed in opengl vertex array */
    struct uncolored_space_triangle_payload {
        t_fvec3 v0;
        t_fvec2 uv0;
        t_fvec3 v1;
        t_fvec2 uv1;
        t_fvec3 v2;
        t_fvec2 uv2;
    } payload;
};

/* batch of primitives with overlapping properties */
struct mesh_batch {
    GLuint buffer;          /* server side storage */
    size_t buffer_len;      /* element count */
    uint8_t *primitives;
};

struct mesh_batch_item {
    t_texture_key key;
    struct mesh_batch value;
};

#endif
